@@ -1,7 +1,7 @@
 /*!
 A crate for parsing Bib(La)TeX files.
 
-The main API entrypoint is a [`Bibliography`].
+The main API entrypoint is the [`Bibliography`] struct.
 
 # Example
 
@@ -26,18 +26,18 @@ mod raw;
 mod resolve;
 mod types;
 
-pub use chunk::{Chunk, Chunks, ChunksExt};
-use macros::*;
+pub use chunk::{Chunk, Chunks, ChunksExt, ChunksRef};
 pub use mechanics::EntryType;
-pub use raw::{RawBibliography, RawEntry};
+pub use raw::{ParseError, ParseErrorKind, Token};
 pub use types::*;
 
 use std::collections::{BTreeMap, HashMap};
-use std::error::Error;
 use std::fmt;
-use std::fmt::{Display, Formatter, Write};
+use std::fmt::{Debug, Display, Formatter, Write};
 
-use mechanics::{AuthorMode, PagesChapterMode};
+use macros::*;
+use mechanics::{is_verbatim_field, AuthorMode, PagesChapterMode};
+use raw::RawBibliography;
 
 use paste::paste;
 
@@ -67,6 +67,38 @@ pub struct Entry {
     pub fields: BTreeMap<String, Chunks>,
 }
 
+/// Errors that can occur when retrieving a field of an [`Entry`].
+#[derive(Debug, Clone)]
+pub enum RetrievalError {
+    /// The entry has no field with this name.
+    Missing(String),
+    /// The field contains malformed data.
+    TypeError(TypeError),
+}
+
+impl Display for RetrievalError {
+    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
+        match self {
+            Self::Missing(s) => write!(f, "field {} is missing", s),
+            Self::TypeError(err) => write!(f, "{}", err),
+        }
+    }
+}
+
+impl From<TypeError> for RetrievalError {
+    fn from(err: TypeError) -> Self {
+        Self::TypeError(err)
+    }
+}
+
+fn convert_result<T>(err: Result<T, RetrievalError>) -> Result<Option<T>, TypeError> {
+    match err {
+        Ok(val) => Ok(Some(val)),
+        Err(RetrievalError::Missing(_)) => Ok(None),
+        Err(RetrievalError::TypeError(err)) => Err(err),
+    }
+}
+
 impl Bibliography {
     /// Create a new, empty bibliography.
     pub fn new() -> Self {
@@ -77,41 +109,46 @@ impl Bibliography {
     }
 
     /// Parse a bibliography from a source string.
-    pub fn parse(src: &str) -> Result<Self, BibliographyError> {
-        Self::from_raw(RawBibliography::parse(src))
+    pub fn parse(src: &str) -> Result<Self, ParseError> {
+        Self::from_raw(RawBibliography::parse(src)?)
     }
 
-    /// Construct a bibliography from a raw bibliography.
-    pub fn from_raw(raw: RawBibliography) -> Result<Self, BibliographyError> {
+    /// Construct a bibliography from a raw bibliography, with the `xdata` and
+    /// `crossref` links resolved.
+    fn from_raw(raw: RawBibliography) -> Result<Self, ParseError> {
         let mut res = Self::new();
         let abbr = &raw.abbreviations;
 
         for entry in raw.entries {
             // Check that the key is not repeated
-            if res.get(entry.key).is_some() {
-                return Err(BibliographyError::DuplicateKey(entry.key.to_string()));
+            if res.get(entry.v.key).is_some() {
+                return Err(ParseError::new(
+                    entry.span,
+                    ParseErrorKind::DuplicateKey(entry.v.key.to_string()),
+                ));
             }
 
             let mut fields = BTreeMap::new();
-            for (field_key, field_value) in entry.fields.into_iter() {
+            for (field_key, field_value) in entry.v.fields.into_iter() {
                 let field_key = field_key.to_string();
-
-                if let Some(r) = resolve::resolve(field_value, abbr) {
-                    fields.insert(field_key, r);
-                } else {
-                    return Err(BibliographyError::MalformedField(
-                        entry.key.to_string(),
-                        field_key,
-                    ));
-                }
+                let parsed = resolve::parse_field(&field_key, &field_value, abbr)?;
+                fields.insert(field_key, parsed);
             }
             res.insert(Entry {
-                key: entry.key.to_string(),
-                entry_type: EntryType::new(entry.entry_type),
+                key: entry.v.key.to_string(),
+                entry_type: EntryType::new(entry.v.kind),
                 fields,
             });
         }
 
+        let mut entries = res.entries.clone();
+        for entry in &mut entries {
+            entry.resolve_crossrefs(&res).map_err(|e| {
+                ParseError::new(e.span, ParseErrorKind::ResolutionError(e.kind))
+            })?;
+        }
+        res.entries = entries;
+
         Ok(res)
     }
 
@@ -132,14 +169,6 @@ impl Bibliography {
         self.entries.get_mut(index)
     }
 
-    /// Try to find an entry with the given cite key and return a copy of that
-    /// entry with all `crossref` and `xdata` dependencies resolved.
-    pub fn get_resolved(&self, key: &str) -> Option<Entry> {
-        let mut entry = self.get(key)?.clone();
-        entry.resolve_crossrefs(self);
-        Some(entry)
-    }
-
     /// Insert an entry into the bibliography.
     ///
     /// If an entry with the same cite key is already present, the entry is
@@ -150,7 +179,8 @@ impl Bibliography {
         } else {
             let index = self.entries.len();
             self.keys.insert(entry.key.clone(), index);
-            if let Some(ids) = entry.get_as::<Vec<String>>("ids") {
+            if let Some(ids) = convert_result(entry.get_as::<Vec<String>>("ids")).unwrap()
+            {
                 for alias in ids {
                     self.keys.insert(alias, index);
                 }
@@ -234,7 +264,11 @@ impl Bibliography {
             if !first {
                 write!(sink, "\n")?;
             }
-            writeln!(sink, "{}", entry.to_bibtex_string())?;
+            writeln!(
+                sink,
+                "{}",
+                entry.to_bibtex_string().map_err(|_| fmt::Error)?
+            )?;
             first = false;
         }
         Ok(())
@@ -257,29 +291,6 @@ impl IntoIterator for Bibliography {
     }
 }
 
-/// Errors that may occur when converting a [`RawBibliography`] into a [`Bibliography`]
-#[derive(Debug, Clone, PartialEq)]
-#[non_exhaustive]
-pub enum BibliographyError {
-    /// The key already occurred.
-    DuplicateKey(String),
-    /// The key contains a malformed field.
-    MalformedField(String, String),
-}
-
-impl Display for BibliographyError {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        match self {
-            Self::DuplicateKey(key) => write!(f, "Duplicate key \"{}\"", key),
-            Self::MalformedField(key, field) => {
-                write!(f, "Key \"{}\" contains malformed field \"{}\"", key, field)
-            }
-        }
-    }
-}
-
-impl Error for BibliographyError {}
-
 impl Entry {
     /// Construct new, empty entry.
     pub fn new(key: String, entry_type: EntryType) -> Self {
@@ -289,15 +300,18 @@ impl Entry {
     /// Get the chunk slice of a field.
     ///
     /// The field key must be lowercase.
-    pub fn get(&self, key: &str) -> Option<&[Chunk]> {
+    pub fn get(&self, key: &str) -> Option<ChunksRef> {
         self.fields.get(key).map(AsRef::as_ref)
     }
 
     /// Parse the value of a field into a specific type.
     ///
     /// The field key must be lowercase.
-    pub fn get_as<T: Type>(&self, key: &str) -> Option<T> {
-        self.get(key)?.parse::<T>()
+    pub fn get_as<T: Type>(&self, key: &str) -> Result<T, RetrievalError> {
+        self.get(key)
+            .ok_or_else(|| RetrievalError::Missing(key.to_string()))?
+            .parse::<T>()
+            .map_err(Into::into)
     }
 
     /// Set the chunk slice for a field.
@@ -320,34 +334,25 @@ impl Entry {
     }
 
     /// The parents of an entry in a semantic sense (`crossref` and `xref`).
-    pub fn parents(&self) -> Vec<String> {
+    pub fn parents(&self) -> Result<Vec<String>, TypeError> {
         let mut parents = vec![];
 
-        if let Some(crossref) = self.get_as::<String>("crossref") {
+        if let Some(crossref) = convert_result(self.get_as::<String>("crossref"))? {
             parents.push(crossref);
         }
 
-        if let Some(xrefs) = self.get_as::<Vec<String>>("xref") {
+        if let Some(xrefs) = convert_result(self.get_as::<Vec<String>>("xref"))? {
             parents.extend(xrefs);
         }
 
-        parents
+        Ok(parents)
     }
 
     /// Verify if the entry has the appropriate fields for its [`EntryType`].
-    ///
-    /// This function returns two vectors: The first indicating fields that
-    /// should have been present but were not, the second indicating fields that
-    /// were set but are forbidden by the [`EntryType`]. Consequently, the entry
-    /// is well-formed if both vectors are empty.
-    ///
-    /// _Note:_ This function will not resolve the entry. If you want to support
-    /// / expect files using `crossref` and `xdata`, you will want to call this
-    /// method only on entries obtained through [`Bibliography::get_resolved`].
-    pub fn verify(&self) -> (Vec<&str>, Vec<&str>) {
+    pub fn verify(&self) -> Report {
         let reqs = self.entry_type.requirements();
         let mut missing = vec![];
-        let mut outlawed = vec![];
+        let mut superfluous = vec![];
 
         for field in reqs.required {
             match field {
@@ -392,35 +397,36 @@ impl Entry {
 
         for field in reqs.forbidden {
             if self.get_non_empty(field).is_some() {
-                outlawed.push(field);
+                superfluous.push(field);
             }
         }
 
         match reqs.author_eds_field {
             AuthorMode::OneRequired => {
-                if self.author().is_none() && self.editors().is_empty() {
+                if self.author().is_err() && self.editors().unwrap_or_default().is_empty()
+                {
                     missing.push("author");
                 }
             }
             AuthorMode::BothRequired => {
-                if self.editors().is_empty() {
+                if self.editors().unwrap_or_default().is_empty() {
                     missing.push("editor");
                 }
-                if self.author().is_none() {
+                if self.author().is_err() {
                     missing.push("author");
                 }
             }
             AuthorMode::AuthorRequired | AuthorMode::AuthorRequiredEditorOptional => {
-                if self.author().is_none() {
+                if self.author().is_err() {
                     missing.push("author");
                 }
             }
             AuthorMode::EditorRequiredAuthorForbidden => {
-                if self.editors().is_empty() {
+                if self.editors().unwrap_or_default().is_empty() {
                     missing.push("editor");
                 }
-                if self.author().is_some() {
-                    outlawed.push("author");
+                if self.author().is_ok() {
+                    superfluous.push("author");
                 }
             }
             _ => {}
@@ -428,33 +434,73 @@ impl Entry {
 
         match reqs.page_chapter_field {
             PagesChapterMode::OneRequired => {
-                if self.pages().is_none() && self.chapter().is_none() {
+                if self.pages().is_err() && self.chapter().is_err() {
                     missing.push("pages");
                 }
             }
             PagesChapterMode::BothForbidden => {
-                if self.pages().is_some() {
-                    outlawed.push("pages");
+                if self.pages().is_ok() {
+                    superfluous.push("pages");
                 }
-                if self.chapter().is_some() {
-                    outlawed.push("chapter");
+                if self.chapter().is_ok() {
+                    superfluous.push("chapter");
                 }
             }
             PagesChapterMode::PagesRequired => {
-                if self.pages().is_none() {
+                if self.pages().is_err() {
                     missing.push("pages");
                 }
             }
             _ => {}
         }
 
+        let mut malformed = vec![];
+
+        for (key, chunks) in &self.fields {
+            let error = match key.as_str() {
+                "edition" => chunks.parse::<Edition>().err(),
+                "organization" => chunks.parse::<Vec<Chunks>>().err(),
+                "pages" => chunks.parse::<Vec<std::ops::Range<u32>>>().err(),
+                "publisher" => chunks.parse::<Vec<Chunks>>().err(),
+                "volume" => chunks.parse::<i64>().err(),
+                "bookpagination" => chunks.parse::<Pagination>().err(),
+                "pagination" => chunks.parse::<Pagination>().err(),
+                "volumes" => chunks.parse::<i64>().err(),
+                "gender" => chunks.parse::<Gender>().err(),
+                "editortype" => chunks.parse::<EditorType>().err(),
+                "editoratype" => chunks.parse::<EditorType>().err(),
+                "editorbtype" => chunks.parse::<EditorType>().err(),
+                "editorctype" => chunks.parse::<EditorType>().err(),
+                "xref" => chunks.parse::<Vec<String>>().err(),
+                "xdata" => chunks.parse::<Vec<String>>().err(),
+                "ids" => chunks.parse::<Vec<String>>().err(),
+                _ => continue,
+            };
+
+            match error {
+                Some(t) => malformed.push((key.clone(), t)),
+                None => {}
+            }
+        }
+
+        for (key, err) in [
+            ("date", self.date().err()),
+            ("urldate", self.url_date().err()),
+            ("origdate", self.orig_date().err()),
+            ("eventdate", self.event_date().err()),
+        ] {
+            if let Some(RetrievalError::TypeError(t)) = err {
+                malformed.push((key.to_string(), t));
+            }
+        }
+
         if reqs.needs_date {
-            if self.date().is_none() {
+            if let Err(RetrievalError::Missing(_)) = self.date() {
                 missing.push("year");
             }
         }
 
-        (missing, outlawed)
+        Report { missing, superfluous, malformed }
     }
 
     /// Serialize this entry into a BibLaTeX string.
@@ -472,7 +518,13 @@ impl Entry {
                 k => k,
             };
 
-            writeln!(biblatex, "{} = {},", key, value.to_biblatex_string()).unwrap();
+            writeln!(
+                biblatex,
+                "{} = {},",
+                key,
+                value.to_biblatex_string(is_verbatim_field(key))
+            )
+            .unwrap();
         }
 
         biblatex.push('}');
@@ -480,7 +532,9 @@ impl Entry {
     }
 
     /// Serialize this entry into a BibTeX string.
-    pub fn to_bibtex_string(&self) -> String {
+    ///
+    /// This function can return an error if there is a malformed date field.
+    pub fn to_bibtex_string(&self) -> Result<String, TypeError> {
         let mut bibtex = String::new();
         let ty = self.entry_type.to_bibtex();
         let thesis = matches!(ty, EntryType::PhdThesis | EntryType::MastersThesis);
@@ -489,9 +543,10 @@ impl Entry {
 
         for (key, value) in &self.fields {
             if key == "date" {
-                if let Some(date) = self.date() {
+                if let Some(date) = convert_result(self.date())? {
                     for (key, value) in date.to_fieldset() {
-                        let v = [Chunk::Normal(value)].to_biblatex_string();
+                        let v = [Spanned::zero(Chunk::Normal(value))]
+                            .to_biblatex_string(false);
                         writeln!(bibtex, "{} = {},", key, v).unwrap();
                     }
                 }
@@ -505,44 +560,51 @@ impl Entry {
                 k => k,
             };
 
-            writeln!(bibtex, "{} = {},", key, value.to_biblatex_string()).unwrap();
+            writeln!(
+                bibtex,
+                "{} = {},",
+                key,
+                value.to_biblatex_string(is_verbatim_field(key))
+            )
+            .unwrap();
         }
 
         bibtex.push('}');
-        bibtex
+        Ok(bibtex)
     }
 
     /// Get an entry but return None for empty chunk slices.
-    fn get_non_empty(&self, key: &str) -> Option<&[Chunk]> {
+    fn get_non_empty(&self, key: &str) -> Option<ChunksRef> {
         let entry = self.get(key)?;
         if !entry.is_empty() { Some(entry) } else { None }
     }
 
     /// Resolves all data dependancies defined by `crossref` and `xdata` fields.
-    fn resolve_crossrefs(&mut self, bib: &Bibliography) {
+    fn resolve_crossrefs(&mut self, bib: &Bibliography) -> Result<(), TypeError> {
         let mut refs = vec![];
 
-        if let Some(crossref) = self.get_as::<String>("crossref") {
+        if let Some(crossref) = convert_result(self.get_as::<String>("crossref"))? {
             refs.extend(bib.get(&crossref).cloned());
         }
 
-        if let Some(keys) = self.get_as::<Vec<String>>("xdata") {
+        if let Some(keys) = convert_result(self.get_as::<Vec<String>>("xdata"))? {
             for key in keys {
                 refs.extend(bib.get(&key).cloned());
             }
         }
 
         for mut crossref in refs {
-            crossref.resolve_crossrefs(bib);
-            self.resolve_single_crossref(crossref);
+            crossref.resolve_crossrefs(bib)?;
+            self.resolve_single_crossref(crossref)?;
         }
 
-        self.remove("crossref");
         self.remove("xdata");
+
+        Ok(())
     }
 
     /// Resolve data dependencies using another entry.
-    fn resolve_single_crossref(&mut self, crossref: Entry) {
+    fn resolve_single_crossref(&mut self, crossref: Entry) -> Result<(), TypeError> {
         let req = self.entry_type.requirements();
 
         let mut relevant = req.required;
@@ -638,21 +700,42 @@ impl Entry {
         }
 
         if self.entry_type == EntryType::XData {
-            return;
+            return Ok(());
         }
 
         if req.needs_date {
-            if let Some(date) = crossref.date() {
+            if let Some(date) = convert_result(crossref.date())? {
                 self.set_date(date);
             }
         }
+
+        Ok(())
+    }
+}
+
+/// A report of the validity of an `Entry`. Can be obtained by calling [`Entry::verify`].
+pub struct Report {
+    /// These fields were missing although they are required for the entry type.
+    pub missing: Vec<&'static str>,
+    /// These fields were present but are not allowed for the entry type.
+    pub superfluous: Vec<&'static str>,
+    /// These fields were present but contained malformed data.
+    pub malformed: Vec<(String, TypeError)>,
+}
+
+impl Report {
+    /// Whether the report is empty and contains no errors.
+    pub fn is_ok(&self) -> bool {
+        self.missing.is_empty()
+            && self.superfluous.is_empty()
+            && self.malformed.is_empty()
     }
 }
 
 impl Entry {
     // BibTeX fields.
     fields! {
-        // Fields without a specified return type simply return `&[Chunk]`.
+        // Fields without a specified return type simply return `ChunksRef`.
         author: "author" => Vec<Person>,
         book_title: "booktitle",
         chapter: "chapter",
@@ -694,25 +777,29 @@ impl Entry {
     /// to four entries, one for each editorial role.
     ///
     /// The default `EditorType::Editor` is assumed if the type field is empty.
-    pub fn editors(&self) -> Vec<(Vec<Person>, EditorType)> {
+    pub fn editors(&self) -> Result<Vec<(Vec<Person>, EditorType)>, TypeError> {
         let mut editors = vec![];
 
-        let mut parse = |name_field: &str, editor_field: &str| {
-            if let Some(persons) = self.get_as::<Vec<Person>>(name_field) {
+        let mut parse = |name_field: &str, editor_field: &str| -> Result<(), TypeError> {
+            if let Some(persons) = convert_result(self.get_as::<Vec<Person>>(name_field))?
+            {
                 let editor_type = self
                     .get(editor_field)
-                    .and_then(|chunks| chunks.parse::<EditorType>())
+                    .map(|chunks| chunks.parse::<EditorType>())
+                    .transpose()?
                     .unwrap_or(EditorType::Editor);
                 editors.push((persons, editor_type));
             }
+
+            Ok(())
         };
 
-        parse("editor", "editortype");
-        parse("editora", "editoratype");
-        parse("editorb", "editorbtype");
-        parse("editorc", "editorctype");
+        parse("editor", "editortype")?;
+        parse("editora", "editoratype")?;
+        parse("editorb", "editorbtype")?;
+        parse("editorc", "editorctype")?;
 
-        editors
+        Ok(editors)
     }
 
     // BibLaTeX supplemental fields.
@@ -784,11 +871,78 @@ impl Entry {
     }
 }
 
+type Span = std::ops::Range<usize>;
+
+/// A value with the span it corresponds to in the source code.
+///
+/// Spans can be _detatched,_ this means that they deliberately do not point
+/// into the source code. Such spans are created when manually setting fields
+/// with an empty bibliography or after parsing a file. Detached spans do not
+/// indicate valid index ranges in the source files and must not be used as
+/// such. A spanned item can be checked for detachment by calling
+/// [`Self::is_detached`].
+#[derive(Clone, Eq, PartialEq)]
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+pub struct Spanned<T> {
+    /// The spanned value.
+    pub v: T,
+    /// The location in source code of the value.
+    pub span: Span,
+}
+
+impl<T> Spanned<T> {
+    /// Create a new instance from a value and its span.
+    pub fn new(v: T, span: Span) -> Self {
+        Self { v, span }
+    }
+
+    /// Create a new instance with a value and a zero-length span.
+    pub fn zero(v: T) -> Self {
+        Self { v, span: 0 .. 0 }
+    }
+
+    /// Create a new instance with a detached span.
+    pub fn detached(v: T) -> Self {
+        Self { v, span: usize::MAX .. usize::MAX }
+    }
+
+    /// Whether the span is detached.
+    pub fn is_detached(&self) -> bool {
+        self.span.start == usize::MAX
+    }
+
+    /// Convert from `&Spanned<T>` to `Spanned<&T>`
+    pub fn as_ref(&self) -> Spanned<&T> {
+        Spanned { v: &self.v, span: self.span.clone() }
+    }
+
+    /// Map the value using a function keeping the span.
+    pub fn map<F, U>(self, f: F) -> Spanned<U>
+    where
+        F: FnOnce(T) -> U,
+    {
+        Spanned { v: f(self.v), span: self.span }
+    }
+}
+
+impl<T: Debug> Debug for Spanned<T> {
+    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
+        self.v.fmt(f)?;
+        if f.alternate() {
+            f.write_str(" <")?;
+            self.span.fmt(f)?;
+            f.write_str(">")?;
+        }
+        Ok(())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use std::fs;
 
     use super::*;
+    use crate::raw::Token;
 
     #[test]
     fn test_correct_bib() {
@@ -804,30 +958,82 @@ mod tests {
         match bibliography {
             Ok(_) => panic!("Should return Err"),
             Err(s) => {
-                assert_eq!(s, BibliographyError::DuplicateKey("ishihara2012".into()));
+                assert_eq!(s.kind, ParseErrorKind::DuplicateKey("ishihara2012".into()));
             }
         };
     }
 
     #[test]
     fn test_parse_incorrect_result() {
-        let contents = fs::read_to_string("tests/incorrect.bib").unwrap();
+        let contents = fs::read_to_string("tests/incorrect_syntax.bib").unwrap();
 
         let bibliography = Bibliography::parse(&contents);
         match bibliography {
-            Ok(_) => panic!("Should return Err"),
+            Ok(_) => {
+                panic!("Should return Err")
+            }
             Err(s) => {
                 assert_eq!(
                     s,
-                    BibliographyError::MalformedField(
-                        "conigliocorbalan".into(),
-                        "author".into()
-                    )
+                    ParseError::new(369 .. 369, ParseErrorKind::Expected(Token::Equals))
                 );
             }
         };
     }
 
+    #[test]
+    fn test_parse_incorrect_types() {
+        let contents = fs::read_to_string("tests/incorrect_data.bib").unwrap();
+
+        let bibliography = Bibliography::parse(&contents).unwrap();
+        let rashid = bibliography.get("rashid2016").unwrap();
+        match rashid.pagination() {
+            Err(RetrievalError::TypeError(s)) => {
+                assert_eq!(
+                    s,
+                    TypeError::new(352 .. 359, TypeErrorKind::UnknownPagination)
+                );
+            }
+            _ => {
+                panic!()
+            }
+        };
+
+        match rashid.date() {
+            Err(RetrievalError::TypeError(s)) => {
+                assert_eq!(
+                    s,
+                    TypeError::new(295 .. 300, TypeErrorKind::WrongNumberOfDigits)
+                );
+            }
+            _ => {
+                panic!()
+            }
+        };
+
+        let coniglio = bibliography.get("conigliocorbalan").unwrap();
+        match coniglio.date() {
+            Err(RetrievalError::TypeError(s)) => {
+                assert_eq!(
+                    s,
+                    TypeError::new(783 .. 785, TypeErrorKind::MonthOutOfRange)
+                );
+            }
+            _ => {
+                panic!()
+            }
+        };
+
+        match coniglio.pages() {
+            Err(RetrievalError::TypeError(s)) => {
+                assert_eq!(s, TypeError::new(759 .. 759, TypeErrorKind::InvalidNumber));
+            }
+            _ => {
+                panic!()
+            }
+        };
+    }
+
     #[test]
     fn test_gral_paper() {
         dump_debug("tests/gral.bib");
@@ -876,7 +1082,8 @@ mod tests {
         let biblatex = bibliography.get_mut("haug2019").unwrap().to_biblatex_string();
         assert!(biblatex.contains("institution = {Technische Universität Berlin},"));
 
-        let bibtex = bibliography.get_mut("haug2019").unwrap().to_bibtex_string();
+        let bibtex =
+            bibliography.get_mut("haug2019").unwrap().to_bibtex_string().unwrap();
         assert!(bibtex.contains("school = {Technische Universität Berlin},"));
         assert!(bibtex.contains("year = {2019},"));
         assert!(bibtex.contains("month = {10},"));
@@ -889,20 +1096,27 @@ mod tests {
         let contents = fs::read_to_string("tests/cross.bib").unwrap();
         let mut bibliography = Bibliography::parse(&contents).unwrap();
 
-        let ok = (vec![], vec![]);
-        assert_eq!(bibliography.get_mut("haug2019").unwrap().verify(), ok);
-        assert_eq!(bibliography.get_mut("cannonfodder").unwrap().verify(), ok);
+        assert_eq!(
+            bibliography.get_mut("haug2019").unwrap().verify().is_ok(),
+            true
+        );
+        assert_eq!(
+            bibliography.get_mut("cannonfodder").unwrap().verify().is_ok(),
+            true
+        );
 
         let ill = bibliography.get("ill-defined").unwrap();
-        let (missing, outlawed) = ill.verify();
-        assert_eq!(missing.len(), 3);
-        assert_eq!(outlawed.len(), 3);
-        assert!(missing.contains(&"title"));
-        assert!(missing.contains(&"year"));
-        assert!(missing.contains(&"editor"));
-        assert!(outlawed.contains(&"maintitle"));
-        assert!(outlawed.contains(&"author"));
-        assert!(outlawed.contains(&"chapter"));
+        let report = ill.verify();
+        assert_eq!(report.missing.len(), 3);
+        assert_eq!(report.superfluous.len(), 3);
+        assert_eq!(report.malformed.len(), 1);
+        assert!(report.missing.contains(&"title"));
+        assert!(report.missing.contains(&"year"));
+        assert!(report.missing.contains(&"editor"));
+        assert!(report.superfluous.contains(&"maintitle"));
+        assert!(report.superfluous.contains(&"author"));
+        assert!(report.superfluous.contains(&"chapter"));
+        assert_eq!(report.malformed[0].0.as_str(), "gender");
     }
 
     #[test]
@@ -910,14 +1124,14 @@ mod tests {
         let contents = fs::read_to_string("tests/cross.bib").unwrap();
         let bibliography = Bibliography::parse(&contents).unwrap();
 
-        let e = bibliography.get_resolved("macmillan").unwrap();
+        let e = bibliography.get("macmillan").unwrap();
         assert_eq!(e.publisher().unwrap()[0].format_verbatim(), "Macmillan");
         assert_eq!(
             e.location().unwrap().format_verbatim(),
             "New York and London"
         );
 
-        let book = bibliography.get_resolved("recursive").unwrap();
+        let book = bibliography.get("recursive").unwrap();
         assert_eq!(book.publisher().unwrap()[0].format_verbatim(), "Macmillan");
         assert_eq!(
             book.location().unwrap().format_verbatim(),
@@ -928,13 +1142,13 @@ mod tests {
             "Recursive shennenigans and other important stuff"
         );
 
-        assert_eq!(bibliography.get("arrgh").unwrap().parents(), vec![
+        assert_eq!(bibliography.get("arrgh").unwrap().parents().unwrap(), vec![
             "polecon".to_string()
         ]);
-        let arrgh = bibliography.get_resolved("arrgh").unwrap();
+        let arrgh = bibliography.get("arrgh").unwrap();
         assert_eq!(arrgh.entry_type, EntryType::Article);
         assert_eq!(arrgh.volume().unwrap(), 115);
-        assert_eq!(arrgh.editors()[0].0[0].name, "Uhlig");
+        assert_eq!(arrgh.editors().unwrap()[0].0[0].name, "Uhlig");
         assert_eq!(arrgh.number().unwrap().format_verbatim(), "6");
         assert_eq!(
             arrgh.journal().unwrap().format_verbatim(),
@@ -966,4 +1180,28 @@ mod tests {
             println!("\"{}\".", x.title().unwrap().format_sentence());
         }
     }
+
+    #[test]
+    fn test_verbatim_fields() {
+        let contents = fs::read_to_string("tests/libra.bib").unwrap();
+        let bibliography = Bibliography::parse(&contents).unwrap();
+
+        // Import an entry/field with escaped colons
+        let e = bibliography.get("dierksmeierJustHODLMoral2018").unwrap();
+        assert_eq!(e.doi().unwrap(), "10.1007/s41463-018-0036-z");
+        assert_eq!(
+            e.file().unwrap(),
+            "C:\\Users\\mhaug\\Zotero\\storage\\DTPR7TES\\Dierksmeier - 2018 - Just HODL On the Moral Claims of Bitcoin and Ripp.pdf"
+        );
+
+        // Import an entry/field with unescaped colons
+        let e = bibliography.get("LibraAssociationIndependent").unwrap();
+        assert_eq!(e.url().unwrap(), "https://libra.org/association/");
+
+        // Test export of entry (not escaping colons)
+        let e = bibliography.get("finextraFedGovernorChallenges2019").unwrap();
+        assert_eq!(e.to_biblatex_string(), 
+        "@online{finextraFedGovernorChallenges2019,\nauthor = {FinExtra},\ndate = {2019-12-18},\nfile = {C:\\\\Users\\\\mhaug\\\\Zotero\\\\storage\\\\VY9LAKFE\\\\fed-governor-challenges-facebooks-libra-project.html},\ntitle = {Fed {Governor} Challenges {Facebook}'s {Libra} Project},\nurl = {https://www.finextra.com/newsarticle/34986/fed-governor-challenges-facebooks-libra-project},\nurldate = {2020-08-22},\n}"
+        )
+    }
 }
